@@ -3,7 +3,7 @@
  * mapping-daemon.c: central daemon to keep track of file mappings
  *
  * Copyright (C) 2002 Red Hat, Inc.
- * Copyright (C) 2005 William Jon McCann <mccann@jhu.edu>
+ * Copyright (C) 2005-2006 William Jon McCann <mccann@jhu.edu>
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -56,6 +56,10 @@ time_t      last_disconnect = 0;
 
 #define KEEP_DATA_TIME 60*60*2
 
+/*
+#define MEMLEAK_DEBUG_TIME 120
+*/
+
 #undef DEBUG_ENABLE
 
 #ifdef DEBUG_ENABLE
@@ -75,7 +79,7 @@ time_t      last_disconnect = 0;
 typedef struct {
 	MappingProtocolChannel *source;
 	GList                  *subscriptions;
-	guint                   ref_count;
+	volatile gint           ref_count;
 } MappingConnection;
 
 typedef struct {
@@ -84,7 +88,7 @@ typedef struct {
 	char              *path;
 	gboolean           cancelled;
 	void              *userdata;
-	guint              ref_count;
+	volatile gint      ref_count;
 } MappingSubscription;
 
 typedef enum {
@@ -104,18 +108,22 @@ typedef struct {
 
 	/* for directories: */
 	GList          *children;
+	volatile gint   ref_count;
 } VirtualNode;
 
 typedef struct {
 	char        *method;
 	char        *tempdir;
 	VirtualNode *root_node;
+	volatile gint ref_count;
 } VirtualRoot;
 
-static GnomeVFSResult monitor_cancel (const char        *rootname,
-				      const char        *path,
-				      MappingConnection *connection,
-				      void              *userdata);
+static GnomeVFSResult monitor_cancel     (const char        *rootname,
+					  const char        *path,
+					  MappingConnection *connection,
+					  void              *userdata);
+static VirtualNode *  virtual_node_ref   (VirtualNode       *node);
+static void           virtual_node_unref (VirtualNode       *node);
 
 #ifdef DEBUG_ENABLE
 static FILE *debug_out = NULL;
@@ -146,6 +154,7 @@ debug_print (const char *format, ...)
 typedef struct {
 	gint32               type;
 	char                *filename;
+	VirtualNode         *node;
 	MappingSubscription *subscription;
 } DispatchData;
 
@@ -162,13 +171,15 @@ subscription_free (MappingSubscription *sub)
 	sub = NULL;
 }
 
-static void
+static MappingSubscription *
 subscription_ref (MappingSubscription *subscription)
 {
-	g_return_if_fail (subscription != NULL);
-	g_return_if_fail (subscription->ref_count > 0);
+	g_return_val_if_fail (subscription != NULL, NULL);
+	g_return_val_if_fail (subscription->ref_count > 0, NULL);
 
-	subscription->ref_count += 1;
+	g_atomic_int_inc (&subscription->ref_count);
+
+	return subscription;
 }
 
 static void
@@ -177,10 +188,9 @@ subscription_unref (MappingSubscription *subscription)
 	g_return_if_fail (subscription != NULL);
 	g_return_if_fail (subscription->ref_count > 0);
 
-	if (subscription->ref_count > 1)
-		subscription->ref_count -= 1;
-	else
+	if (g_atomic_int_dec_and_test (&subscription->ref_count)) {
 		subscription_free (subscription);
+	}
 }
 
 static MappingSubscription *
@@ -234,24 +244,26 @@ subscription_send_event (MappingSubscription *sub,
 			 gint32               type)
 {
 	MappingConnection           *connection;
+	MappingProtocolMessage      *message;
 	MappingProtocolMonitorEvent *event;
-	int                          res;
+	gboolean                     res;
 
 	g_return_val_if_fail (sub != NULL, FALSE);
 	g_return_val_if_fail (filename != NULL, FALSE);
 
-	if (sub->cancelled)
+	if (sub->cancelled) {
+		DEBUG_PRINT ("Skipping event for cancelled subscription: %p %s %d\n", sub, filename, type);
+
 		return FALSE;
+	}
 
 	connection = sub->connection;
 
 	g_return_val_if_fail (connection != NULL, FALSE);
 
-	/* if there is data to be read then delay sending event */
-	if (mapping_protocol_data_available (connection->source))
-		return TRUE;
+	message = mapping_protocol_message_new_monitor_event ();
+	event = MAPPING_PROTOCOL_MONITOR_EVENT (message);
 
-	event = g_new0 (MappingProtocolMonitorEvent, 1);
 	event->type = type;
 	event->path = g_strdup (filename);
 	event->userdata = sub->userdata;
@@ -259,13 +271,9 @@ subscription_send_event (MappingSubscription *sub,
 	DEBUG_PRINT ("Encoding event type %d userdata %p for %s sub: %p conn: %p\n",
 		     type, sub->userdata, filename, sub, connection);
 
-	res = mapping_protocol_monitor_event_encode (connection->source, event);
-	if (res == -1) {
-		subscription_cancel (sub);
-	}
+	res = mapping_protocol_channel_send (connection->source, message);
 
-	g_free (event->path);
-	g_free (event);
+	mapping_protocol_message_unref (message);
 
 	return FALSE;
 }
@@ -286,6 +294,7 @@ dispatch_data_free (gpointer data)
 {
 	DispatchData *ddata = data;
 
+	virtual_node_unref (ddata->node);
 	subscription_unref (ddata->subscription);
 	g_free (ddata->filename);
 	ddata->filename = NULL;
@@ -325,17 +334,21 @@ virtual_node_subscription_event (VirtualNode *node,
 	g_return_if_fail (node != NULL);
 	g_return_if_fail (filename != NULL);
 
+	DEBUG_PRINT ("subscription event for node %p %s %d\n", node, filename, type);
+
 	for (l = node->subscriptions; l != NULL; l = l->next) {
 		int                  timeout;
 		DispatchData        *ddata;
 		MappingSubscription *sub = l->data;
 
-		if (!sub || sub->cancelled)
+		if (!sub || sub->cancelled) {
+			DEBUG_PRINT ("subscription cancelled, skipping\n");
 			continue;
+		}
 
 		ddata = g_new (DispatchData, 1);
-		ddata->subscription = sub;
-		subscription_ref (sub);
+		ddata->node = virtual_node_ref (node);
+		ddata->subscription = subscription_ref (sub);
 		ddata->filename = g_strdup (filename);
 		ddata->type = type;
 
@@ -355,6 +368,7 @@ virtual_node_new (const char     *filename,
 	node = g_new0 (VirtualNode, 1);
 	node->filename = g_strdup (filename);
 	node->type = type;
+	node->ref_count = 1;
 
 	return node;
 }
@@ -365,10 +379,13 @@ virtual_node_free (VirtualNode *node,
 {
 	GList *l;
 
+	DEBUG_PRINT ("freeing node %p %s\n", node, node->filename);
+
 	g_free (node->filename);
+
 	switch (node->type) {
 	case VIRTUAL_NODE_FILE:
-		if (node->backing_file) {
+		if (node->backing_file != NULL) {
 			if (node->owned_file) {
 				g_unlink (node->backing_file);
 			}
@@ -378,10 +395,13 @@ virtual_node_free (VirtualNode *node,
 	case VIRTUAL_NODE_DIRECTORY:
 		if (deep) {
 			for (l = node->children; l != NULL; l = l->next) {
-				virtual_node_free ((VirtualNode *)l->data, TRUE);
+				virtual_node_unref ((VirtualNode *)l->data);
 			}
 		}
 		break;
+	default:
+		g_assert_not_reached ();
+		break;
 	}
 
 	while ((l = g_list_first (node->subscriptions)) != NULL) {
@@ -392,15 +412,60 @@ virtual_node_free (VirtualNode *node,
 	g_free (node);
 }
 
+static void
+virtual_node_unref (VirtualNode *node)
+{
+	g_return_if_fail (node != NULL);
+	g_return_if_fail (node->ref_count > 0);
+
+	if (g_atomic_int_dec_and_test (&node->ref_count)) {
+		virtual_node_free (node, TRUE);
+	}
+}
+
+static VirtualNode *
+virtual_node_ref (VirtualNode *node)
+{
+	g_return_val_if_fail (node != NULL, NULL);
+	g_return_val_if_fail (node->ref_count > 0, NULL);
+
+	g_atomic_int_inc (&node->ref_count);
+	return node;
+}
+
 static void
 virtual_root_free (VirtualRoot *root)
 {
-	virtual_node_free (root->root_node, TRUE);
+	virtual_node_unref (root->root_node);
 	g_free (root->method);
 	g_rmdir (root->tempdir);
 	g_free (root->tempdir);
+	g_free (root);
 }
 
+static void
+virtual_root_unref (VirtualRoot *node)
+{
+	g_return_if_fail (node != NULL);
+	g_return_if_fail (node->ref_count > 0);
+
+	if (g_atomic_int_dec_and_test (&node->ref_count)) {
+		virtual_root_free (node);
+	}
+}
+
+#if 0
+static VirtualRoot *
+virtual_root_ref (VirtualRoot *node)
+{
+	g_return_val_if_fail (node != NULL, NULL);
+	g_return_val_if_fail (node->ref_count > 0, NULL);
+
+	g_atomic_int_inc (&node->ref_count);
+	return node;
+}
+#endif
+
 static VirtualRoot *
 virtual_root_new (const char *method)
 {
@@ -412,7 +477,7 @@ virtual_root_new (const char *method)
 	tempdir = g_build_filename (g_get_tmp_dir (), filename, NULL);
 	g_free (filename);
 
-	root = g_new (VirtualRoot, 1);
+	root = g_new0 (VirtualRoot, 1);
 
 	dir = mkdtemp (tempdir);
 
@@ -423,6 +488,7 @@ virtual_root_new (const char *method)
 		return NULL;
 	}
 
+	root->ref_count = 1;
 	root->method = g_strdup (method);
 	root->tempdir = g_strdup (dir);
 	root->root_node = virtual_node_new (NULL, VIRTUAL_NODE_DIRECTORY);
@@ -535,6 +601,8 @@ virtual_mkdir (VirtualNode *node,
 	}
 
 	subdir = virtual_node_new (name, VIRTUAL_NODE_DIRECTORY);
+
+	/* list takes ownership of ref */
 	node->children = g_list_append (node->children, subdir);
 
 	return subdir;
@@ -547,6 +615,7 @@ virtual_unlink (VirtualNode *dir,
 	g_assert (dir->type == VIRTUAL_NODE_DIRECTORY);
 
 	dir->children = g_list_remove (dir->children, node);
+	virtual_node_unref (node);
 }
 
 static VirtualNode *
@@ -577,7 +646,7 @@ virtual_create (VirtualRoot *root,
 
 		if (fd < 0) {
 			g_free (template);
-			virtual_node_free (file, FALSE);
+			virtual_node_unref (file);
 			return NULL;
 		}
 		close (fd);
@@ -587,6 +656,7 @@ virtual_create (VirtualRoot *root,
 		file->owned_file = TRUE;
 	}
 
+	/* list takes ownership of ref */
 	dir->children = g_list_append (dir->children, file);
 
 	return file;
@@ -683,12 +753,16 @@ remove_dir (const char *rootname,
 		return GNOME_VFS_ERROR_DIRECTORY_NOT_EMPTY;
 	}
 
+	/* temp ref */
+	virtual_node_ref (file);
+
 	virtual_unlink (dir, file);
 
 	virtual_node_subscription_event (file, path, GNOME_VFS_MONITOR_EVENT_DELETED);
 	virtual_node_subscription_event (dir, path, GNOME_VFS_MONITOR_EVENT_DELETED);
 
-	virtual_node_free (file, FALSE);
+	/* drop temp ref */
+	virtual_node_unref (file);
 
 	return GNOME_VFS_OK;
 }
@@ -712,12 +786,16 @@ remove_file (const char *rootname,
 		return GNOME_VFS_ERROR_IS_DIRECTORY;
 	}
 
+	/* temp ref */
+	virtual_node_ref (file);
+
 	virtual_unlink (dir, file);
 
 	virtual_node_subscription_event (file, path, GNOME_VFS_MONITOR_EVENT_DELETED);
 	virtual_node_subscription_event (dir, path, GNOME_VFS_MONITOR_EVENT_DELETED);
 
-	virtual_node_free (file, FALSE);
+	/* drop temp ref */
+	virtual_node_unref (file);
 
 	return GNOME_VFS_OK;
 }
@@ -800,6 +878,7 @@ create_dir_link (VirtualRoot *root,
 	}
 
 	while ((d_name = g_dir_read_name (d)) != NULL) {
+		VirtualNode *file;
 
 		path = g_build_filename (target_path, d_name, NULL);
 
@@ -807,7 +886,7 @@ create_dir_link (VirtualRoot *root,
 			if (S_ISDIR (stat_buf.st_mode)) {
 				create_dir_link (root, dir, d_name, path);
 			} else {
-				virtual_create (root, dir, d_name, path);
+				file = virtual_create (root, dir, d_name, path);
 			}
 		}
 		g_free (path);
@@ -899,10 +978,17 @@ move_file (const char *rootname,
 		}
 		g_free (old_node->filename);
 		old_node->filename = g_strdup (new_node->filename);
+
+		/* temp ref */
+		virtual_node_ref (old_node);
+
 		virtual_unlink (new_dir, new_node);
 		virtual_unlink (old_dir, old_node);
-		new_dir->children = g_list_append (new_dir->children, old_node);
-		virtual_node_free (new_node, FALSE);
+
+		new_dir->children = g_list_append (new_dir->children, virtual_node_ref (old_node));
+
+		/* drop temp ref */
+		virtual_node_unref (old_node);
 	} else {
 		dirname = g_path_get_dirname (new_path);
 		new_dir = virtual_node_lookup (root, dirname, NULL);
@@ -912,8 +998,16 @@ move_file (const char *rootname,
 		}
 		g_free (old_node->filename);
 		old_node->filename = g_path_get_basename (new_path);
+
+		/* temp ref */
+		virtual_node_ref (old_node);
+
 		virtual_unlink (old_dir, old_node);
-		new_dir->children = g_list_append (new_dir->children, old_node);
+
+		new_dir->children = g_list_append (new_dir->children, virtual_node_ref (old_node));
+
+		/* drop temp ref */
+		virtual_node_unref (old_node);
 	}
 
 	virtual_node_subscription_event (new_dir, new_path, GNOME_VFS_MONITOR_EVENT_CREATED);
@@ -928,7 +1022,7 @@ static GnomeVFSResult
 list_dir (const char *rootname,
 	  const char *path,
 	  gint32     *n_elements,
-	   char    ***listing_out)
+	  char     ***listing_out)
 {
 	VirtualRoot *root;
 	VirtualNode *node;
@@ -953,14 +1047,14 @@ list_dir (const char *rootname,
 	*listing_out = listing;
 	*n_elements = (gint32) len * 2;
 
-	for (i=0, l = node->children; l != NULL; l = l->next, i++) {
+	for (i = 0, l = node->children; l != NULL; l = l->next, i++) {
 		node = (VirtualNode *)l->data;
 
-		listing[i*2] = node->filename;
+		listing [i * 2] = g_strdup (node->filename);
 		if (node->type == VIRTUAL_NODE_FILE) {
-			listing [i*2 + 1] = node->backing_file;
+			listing [i * 2 + 1] = g_strdup (node->backing_file);
 		} else {
-			listing [i*2 + 1] = NULL;
+			listing [i * 2 + 1] = NULL;
 		}
 	}
 
@@ -1041,13 +1135,15 @@ connection_free (MappingConnection *connection)
 	connection = NULL;
 }
 
-static void
+static MappingConnection *
 connection_ref (MappingConnection *connection)
 {
-	g_return_if_fail (connection != NULL);
-	g_return_if_fail (connection->ref_count > 0);
+	g_return_val_if_fail (connection != NULL, NULL);
+	g_return_val_if_fail (connection->ref_count > 0, NULL);
 
-	connection->ref_count += 1;
+	g_atomic_int_inc (&connection->ref_count);
+
+	return connection;
 }
 
 static void
@@ -1056,10 +1152,9 @@ connection_unref (MappingConnection *connection)
 	g_return_if_fail (connection != NULL);
 	g_return_if_fail (connection->ref_count > 0);
 
-	if (connection->ref_count > 1)
-		connection->ref_count -= 1;
-	else
+	if (g_atomic_int_dec_and_test (&connection->ref_count)) {
 		connection_free (connection);
+	}
 }
 
 static MappingConnection *
@@ -1123,9 +1218,7 @@ connection_remove (MappingConnection *connection)
 }
 
 static gboolean
-connection_error (MappingConnection *connection,
-		  GIOCondition       condition,
-		  gpointer           data)
+connection_error (MappingConnection *connection)
 {
 	DEBUG_PRINT ("connection_error: source %p\n", connection->source);
 
@@ -1235,110 +1328,100 @@ operation_string (gint32 op)
 	return str;
 }
 
-static gboolean
-handle_request (GIOChannel   *source,
-		GIOCondition  condition,
-		gpointer      data)
+static void
+handle_message (MappingProtocolChannel *channel,
+		MappingProtocolMessage *message,
+		gpointer                data)
 {
-	MappingConnection     *connection;
-	int                    res;
-	MappingProtocolRequest req;
-	MappingProtocolReply   reply;
+	MappingConnection      *connection;
+	MappingProtocolMessage *reply_message;
+	MappingProtocolReply   *reply;
+	MappingProtocolRequest *req;
+	gboolean                res;
 
 	connection = (MappingConnection *)data;
 
-	if ((condition & G_IO_ERR) || (condition & G_IO_HUP)) {
-		connection_error (connection, condition, NULL);
-		return FALSE;
+	if (message == NULL) {
+		connection_error (connection);
+		return;
 	}
 
-	res = mapping_protocol_request_decode (connection->source, &req);
-
-	if (res != 0) {
-		connection_error (connection, condition, NULL);
-		return FALSE;
+	if (message->type != MAPPING_PROTOCOL_MESSAGE_REQUEST) {
+		return;
 	}
 
-	memset (&reply, 0, sizeof (MappingProtocolReply));
+	req = MAPPING_PROTOCOL_REQUEST (message);
 
-	DEBUG_PRINT ("Decoded new request type %s\n", operation_string (req.operation));
+	reply_message = mapping_protocol_message_new_reply (message);
+	reply = MAPPING_PROTOCOL_REPLY (reply_message);
 
-	switch (req.operation) {
+	DEBUG_PRINT ("Decoded new request type %s\n", operation_string (req->operation));
+
+	switch (req->operation) {
 	case MAPPING_PROTOCOL_OP_GET_BACKING_FILE:
-		reply.result = get_backing_file (req.root,
-						 req.path1,
-						 req.option,
-						 &reply.path);
+		reply->result = get_backing_file (req->root,
+						  req->path1,
+						  req->option,
+						  &reply->path);
 		break;
 	case MAPPING_PROTOCOL_OP_LIST_DIR:
-		reply.result = list_dir (req.root,
-					 req.path1,
-					 &reply.n_strings,
-					 &reply.strings);
+		reply->result = list_dir (req->root,
+					  req->path1,
+					  &reply->n_strings,
+					  &reply->strings);
 		break;
 	case MAPPING_PROTOCOL_OP_CREATE_DIR:
-		reply.result = create_dir (req.root,
-					   req.path1);
+		reply->result = create_dir (req->root,
+					    req->path1);
 		break;
 	case MAPPING_PROTOCOL_OP_REMOVE_DIR:
-		reply.result = remove_dir (req.root,
-					   req.path1);
+		reply->result = remove_dir (req->root,
+					    req->path1);
 		break;
 	case MAPPING_PROTOCOL_OP_REMOVE_FILE:
-		reply.result = remove_file (req.root,
-					    req.path1);
+		reply->result = remove_file (req->root,
+					     req->path1);
 		break;
 	case MAPPING_PROTOCOL_OP_CREATE_FILE:
-		reply.result = create_file (req.root,
-					    req.path1,
-					    req.option,
-					    &reply.path,
-					    &reply.option);
+		reply->result = create_file (req->root,
+					     req->path1,
+					     req->option,
+					     &reply->path,
+					     &reply->option);
 		break;
 	case MAPPING_PROTOCOL_OP_CREATE_LINK:
-		reply.result = create_link (req.root,
-					    req.path1,
-					    req.path2);
+		reply->result = create_link (req->root,
+					     req->path1,
+					     req->path2);
 		break;
 	case MAPPING_PROTOCOL_OP_MOVE_FILE:
-		reply.result = move_file (req.root,
-					  req.path1,
-					  req.path2);
+		reply->result = move_file (req->root,
+					   req->path1,
+					   req->path2);
 		break;
 	case MAPPING_PROTOCOL_OP_MONITOR_ADD:
-		reply.result = monitor_add (req.root,
-					    req.path1,
-					    connection,
-					    req.userdata);
+		reply->result = monitor_add (req->root,
+					     req->path1,
+					     connection,
+					     req->userdata);
 		break;
 	case MAPPING_PROTOCOL_OP_MONITOR_CANCEL:
-		reply.result = monitor_cancel (req.root,
-					       req.path1,
-					       connection,
-					       req.userdata);
+		reply->result = monitor_cancel (req->root,
+						req->path1,
+						connection,
+						req->userdata);
 		break;
 	default:
-		g_warning ("Unimplemented op %s\n", operation_string (req.operation));
+		g_warning ("Unimplemented op %s\n", operation_string (req->operation));
 		break;
 	}
 
 	DEBUG_PRINT ("Encoding reply to request op %s path1: %s result: %d path: '%s' n_strings: %d\n",
-		     operation_string (req.operation), req.path1, reply.result, reply.path, reply.n_strings);
+		     operation_string (req->operation), req->path1, reply->result, reply->path, reply->n_strings);
 
-	res = mapping_protocol_reply_encode (connection->source, &reply);
+	res = mapping_protocol_channel_send (channel, reply_message);
 
-	mapping_protocol_request_destroy (&req);
-	mapping_protocol_reply_destroy (&reply);
-
-	if (res != 0) {
-		DEBUG_PRINT ("Could not encode reply: %d\n", res);
-		connection_error (connection, condition, NULL);
-		return FALSE;
-	}
-
-	DEBUG_PRINT ("Finished handling request type %s\n", operation_string (req.operation));
-
-	return TRUE;
+        mapping_protocol_message_unref (reply_message);
 }
 
 static gboolean
@@ -1358,9 +1441,10 @@ handle_new_client (GIOChannel   *source,
         }
 
 	io = mapping_protocol_channel_new (client);
-	connection = connection_new (io);
 
-	g_io_add_watch (io->iochannel, G_IO_IN | G_IO_ERR | G_IO_HUP, handle_request, connection);
+	/* connection takes a reference */
+	connection = connection_new (io);
+	mapping_protocol_channel_set_message_handler (io, handle_message, connection);
 
 	DEBUG_PRINT ("Handling new client source: %p\n", io);
 
@@ -1402,7 +1486,7 @@ free_roots_helper (gpointer       key,
 {
 	VirtualRoot *root = value;
 
-	virtual_root_free (root);
+	virtual_root_unref (root);
 	return TRUE;
 }
 
@@ -1411,6 +1495,8 @@ free_roots (void)
 {
 	g_hash_table_foreach_remove (roots,
 				     free_roots_helper, NULL);
+	g_hash_table_destroy (roots);
+	roots = NULL;
 }
 
 static gboolean
@@ -1433,6 +1519,20 @@ cleanup_timeout (gpointer data)
 	return TRUE;
 }
 
+#ifdef MEMLEAK_DEBUG_TIME
+static gboolean
+debug_shutdown (GMainLoop *loop)
+{
+	g_list_foreach (connections, (GFunc)connection_unref, NULL);
+	g_list_free (connections);
+	connections = NULL;
+
+	g_main_loop_quit (loop);
+
+	return FALSE;
+}
+#endif
+
 int
 main (int   argc,
       char *argv[])
@@ -1497,8 +1597,16 @@ main (int   argc,
 
 	g_timeout_add (5000, &cleanup_timeout, NULL);
 
+#ifdef MEMLEAK_DEBUG_TIME
+        g_timeout_add (MEMLEAK_DEBUG_TIME * 1000,
+                       (GSourceFunc)debug_shutdown,
+                       loop);
+#endif
+
 	g_main_loop_run (loop);
 
+	g_main_loop_unref (loop);
+
 	free_roots ();
 	return 0;
 }
